/**
 * @fileoverview
 * @author Taketshi Aono
 */
'use strict';

var addSingletonGetter = require('node-singleton-getter');
var MultiMap = require('collections/multi-map');
var EventEmitter = require('eventemitter2').EventEmitter2;
var Promise = require('node-promise');
var domain = require('domain');

var exit = process.exit;

/**
 * @param {string} name
 * @param {Object} context
 * @returns {Function}
 */
function getProcessTerminationHandler(name, context, opt_exit) {
  return function processTerminationHandler(code) {
    this._firedMap[name] = true;
    this._events.emit(name);
    Promise.all(this._promises.get(name).concat(this._promises.get('any')))
      .then(exit.bind(process, code || 1));
  }.bind(context);
}


/**
 * @param {string} stack
 * @throws {Error}
 */
function timeoutHandler(stack) {
  console.log('The shutdown handler takes up too much time.\nCREATION STACK :\n' + stack + '\nCURRENT_STACK:\n' + new Error().stack);
  exit.call(process, 1);
};


/**
 * Manage node halting process.
 * @constructor
 */
function ShutdownHandler() {
  this._events = new EventEmitter({delimiter : '::'});
  this._promises = new MultiMap();
  this._domain = domain.create();
  this._firedMap = {};

  //normal
  process.exit = getProcessTerminationHandler('process::exit', this);

  //SIGINT
  process.on('SIGINT', getProcessTerminationHandler('process::interrupted', this));

  //SIGABRT
  process.on('SIGABRT', getProcessTerminationHandler('process::abort', this));

  //SIGTERM
  process.on('SIGTERM', getProcessTerminationHandler('process::terminate', this));

  //uncaughtException
  var domainErrorHandler = getProcessTerminationHandler('domain::error', this);
  this._domain.on('error', function(e) {
    console.log('%s%s', e.message, e.stack);
    domainErrorHandler();
  });

  this._timeout = 3000;
}
addSingletonGetter(ShutdownHandler);


Object.defineProperties(ShutdownHandler.prototype, {
  timeout : {
    /**
     * @param {number} time
     */
    set : function(time) {
      this._timeout = time;
    },

    /**
     * @returns {number}
     */
    get : function() {
      return this._timeout;
    },
    configurable : false,
    enumerable : true
  },

  halting : {
    get : function() {
      return Object.keys(this._firedMap).length > 0;
    }
  }
});


/**
 * @param {string} name
 * @param {Function} fn
 * @param {*} opt_context
 */
ShutdownHandler.prototype.install = function(name, fn, opt_context) {
  if (this._firedMap[name]) {
    return fn.call(opt_context);
  }
  this._addEventHandler(name, fn, opt_context);
};


/**
 * @param {Function} fn
 * @param {*} opt_context
 */
ShutdownHandler.prototype.installAny = function(fn, opt_context) {
  if (Object.keys(this._firedMap).length > 0) {
    return fn.call(opt_context);
  }
  this._addEventHandler('any', fn, opt_context);
};


/**
 * @param {string} name
 * @param {Function} fn
 */
ShutdownHandler.prototype.uninstall = function(name, fn) {
  this._promises.set(name, []);
  this._events.off(name, fn);
};


/**
 * @param {Function} fn
 */
ShutdownHandler.prototype.uninstallAny = function(fn) {
  this._promises.set('any', []);
  this._events.offAny(fn);
};


ShutdownHandler.prototype.runInSandbox = function(fn, opt_context) {
  this._domain.run(function() {fn.call(opt_context)});
};


/**
 * @private
 * @param {string} name
 * @param {Function} fn
 * @param {*} opt_context
 */
ShutdownHandler.prototype._addEventHandler = function(name, fn, opt_context) {
  var defer = Promise.defer();
  var promise = defer.promise;
  var stack = new Error().stack;
  var fired = false;
  this._promises.get(name).push(promise);
  var handler = function() {
        defer.timeout(this._timeout).then(null, timeoutHandler.bind(null, stack));
        if (!fired) {
          fired = true;
          fn.call(opt_context, function() {
            defer.resolve();
          });
        }
      }.bind(this);
  if (name === 'any') {
    this._events.onAny(handler);
  } else {
    this._events.on(name, handler);
  }
};


module.exports = ShutdownHandler;